
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <midifile.h>

int should_shutdown = 0;
int sysex_message_is_sent;
HANDLE signalling_object;

static BOOL WINAPI interrupt_handler(DWORD control_type)
{
	should_shutdown = 1;
	PulseEvent(signalling_object);
	return FALSE;
}

static void CALLBACK midi_out_handler(HMIDIOUT midi_out, UINT message_type, DWORD user_data, DWORD midi_message, DWORD param2)
{
	sysex_message_is_sent = 1;
	PulseEvent(signalling_object);
}

static int get_midi_out_number(char *name)
{
	int midi_out_number, maximum_midi_out_number;

	for (midi_out_number = 0, maximum_midi_out_number = midiOutGetNumDevs(); midi_out_number < maximum_midi_out_number; midi_out_number++)
	{
		MIDIOUTCAPS midi_out_caps;
		midiOutGetDevCaps(midi_out_number, &midi_out_caps, sizeof(midi_out_caps));
		if (strcmp(midi_out_caps.szPname, name) == 0) return midi_out_number;
	}

	if (sscanf(name, "%d", &midi_out_number) == 1) return midi_out_number;
	return -1;
}

static void usage(char *program_name)
{
	fprintf(stderr, "Usage:  %s [ --out <device id> ] [ -- ] <filename.mid>\n", program_name);
	exit(1);
}

int main(int argc, char **argv)
{
	int i;
	int midi_out_number = MIDI_MAPPER;
	char *filename = NULL;
	int double_dash = 0;
	MidiFile_t midi_file;
	HMIDIOUT midi_out;
	unsigned long start_time;
	MidiFileEvent_t midi_file_event;

	for (i = 1; i < argc; i++)
	{
		if (!double_dash && ((strcmp(argv[i], "--help") == 0) || (strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "/help") == 0) || (strcmp(argv[i], "/h") == 0) || (strcmp(argv[i], "/?") == 0)))
		{
			usage(argv[0]);
		}
		else if (!double_dash && (strcmp(argv[i], "--out") == 0))
		{
			if (++i == argc) usage(argv[0]);

			if ((midi_out_number = get_midi_out_number(argv[i])) < 0)
			{
				printf("Error:  No such MIDI output device \"%s\".\n", argv[i]);
				return 1;
			}
		}
		else if (!double_dash && (strcmp(argv[i], "--") == 0))
		{
			double_dash = 1;
		}
		else if (filename == NULL)
		{
			filename = argv[i];
		}
		else
		{
			usage(argv[0]);
		}
	}

	if (filename == NULL)
	{
		usage(argv[0]);
	}

	if ((midi_file = MidiFile_load(filename)) == NULL)
	{
		fprintf(stderr, "Error:  Cannot open \"%s\".\n", filename);
		return 1;
	}

	if (midiOutOpen(&midi_out, midi_out_number, (DWORD)(midi_out_handler), (DWORD)(NULL), CALLBACK_FUNCTION) != MMSYSERR_NOERROR)
	{
		fprintf(stderr, "Error:  Cannot open MIDI output port #%d.\n", midi_out_number);
		return 1;
	}

	signalling_object = CreateEvent(NULL, FALSE, FALSE, NULL);
	SetConsoleCtrlHandler(interrupt_handler, TRUE);
	start_time = timeGetTime();

	for (midi_file_event = MidiFile_getFirstEvent(midi_file); !should_shutdown && (midi_file_event != NULL); midi_file_event = MidiFileEvent_getNextEventInFile(midi_file_event))
	{
		if (MidiFileEvent_getType(midi_file_event) != MIDI_FILE_EVENT_TYPE_META)
		{
			unsigned long event_time = (unsigned long)(MidiFile_getTimeFromTick(midi_file, MidiFileEvent_getTick(midi_file_event)) * 1000);

			while (!should_shutdown)
			{
				unsigned long current_time = timeGetTime();

				if (current_time - start_time < event_time)
				{
					WaitForSingleObject(signalling_object, event_time - (current_time - start_time));
				}
				else
				{
					break;
				}
			}

			if (!should_shutdown)
			{
				if (MidiFileEvent_getType(midi_file_event) == MIDI_FILE_EVENT_TYPE_SYSEX)
				{
					MIDIHDR midi_message_header;
					midi_message_header.lpData = MidiFileSysexEvent_getData(midi_file_event);
					midi_message_header.dwBufferLength = MidiFileSysexEvent_getDataLength(midi_file_event);
					midi_message_header.dwFlags = 0;
					midiOutPrepareHeader(midi_out, &midi_message_header, sizeof(midi_message_header));
					sysex_message_is_sent = 0;
					midiOutLongMsg(midi_out, &midi_message_header, sizeof(midi_message_header));

					while (!sysex_message_is_sent)
					{
						WaitForSingleObject(signalling_object, INFINITE);
					}

					midiOutUnprepareHeader(midi_out, &midi_message_header, sizeof(midi_message_header));
				}
				else
				{
					midiOutShortMsg(midi_out, (DWORD)(MidiFileVoiceEvent_getData(midi_file_event)));
				}
			}
		}
	}

	SetConsoleCtrlHandler(interrupt_handler, FALSE);
	CloseHandle(signalling_object);
	midiOutClose(midi_out);
	MidiFile_free(midi_file);
	return 0;
}

